# MIT License

# Copyright (c) 2024 Andrew Haddad

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import glob
import logging
import os
import shutil
import subprocess
import tarfile

from functools import wraps
from .common import *


def log(logger):
    def middle(func):
        @wraps(func)
        def logged(*args, **kwargs):
            result = func(*args, **kwargs)
            # func_name_fmtd = format_func_name(func=func)
            try:
                inputs = args[0]
            except IndexError:
                inputs = tuple(kwargs.values())[0]
            finally:
                msg_fmtd = f"{func.__module__}:Completed {func.__name__} with input `{inputs}` and output `{result}`"
            logger.info(msg_fmtd)
            return result

        return logged

    return middle


def create_logger(
    name, log_level=logging.INFO, stream=False, file="pharmcat_runner.log"
):
    logger = logging.getLogger(name)
    logger.setLevel(log_level)

    file_handler = logging.FileHandler(file)
    logger.addHandler(file_handler)
    if stream:
        stream_handler = logging.StreamHandler()
        logger.addHandler(stream_handler)
    for handler in logger.handlers:
        handler.setLevel(log_level)
        handler.setFormatter(
            logging.Formatter(LOGGING_FORMAT, datefmt="%Y-%m-%d,%H:%M:%S")
        )
    return logger


def format_func_name(func):
    func_name_fmtd = []
    for i, name in enumerate(func.__name__.split("_")):
        if i == 0:
            if name != "split":
                name = name + "ed"
        func_name_fmtd.append(name)
    return " ".join(func_name_fmtd)


def blockgz_compress(input_file, output_file):
    run_command(f"bgzip -ck {input_file} > {output_file}", shell=True)
    return output_file


def get_files(dir_):
    try:
        files = []
        filenames = os.listdir(dir_)
        for file in filenames:
            files.append(os.path.join(dir_, file))
    except FileNotFoundError:
        files = glob.glob(dir_)
    return files


def validate_environment():
    # Don't need to check for pharmcat or reference genome as these
    # would be installed by the pharmcat_pipeline if they don't already
    # exist
    tools = [
        "bcftools",
        "tabix",
        "plink",
        os.path.join(os.environ.get("JAVA_HOME"), "bin", "java"),
    ]
    for tool in tools:
        try:
            if tool == "bcftools":
                args = [tool, "-v"]
            else:
                args = [tool, "--version"]
            output = run_command(args, capture_output=True)
        except Exception:
            raise RuntimeError(f"Unable to validate existence of {tool}")
        if tool in {"bcftools", "tabix"}:
            version = output.split("\n")[0].split()[-1].split(".")
            version_suffix = int(version[-1])
            if version_suffix < 18:
                raise RuntimeError(
                    f"Version of {tool} must be >= {HTS_LIB_REQUIRED_VERSION}. Found {version}"
                )
    if not os.path.exists(BIN_DIR):
        raise RuntimeError("Unable to find pharmcat installation.")
    for file in os.listdir(BIN_DIR):
        if "preprocessor" in file:
            break
    else:
        raise RuntimeError("Unable to find preprocessor installation.")


def run_command(args, shell=False, capture_output=False):
    stdout = subprocess.run(args, capture_output=True, text=True, shell=shell)
    check_stdout(stdout)
    if capture_output:
        return stdout.stdout


def check_stdout(stdout):
    if stdout.returncode != 0:
        raise RuntimeError(
            f"Error running command {stdout.args} with error: {stdout.stderr}"
        )


def curl(dir_, url):
    run_command(f"cd {dir_} && curl -OL {url}", shell=True)


def unpack_tar(file, dir_="."):
    with tarfile.open(file) as f:
        f.extractall(path=dir_)


def maybe_create_dir(dir_, strict=False):
    if strict:
        try:
            shutil.rmtree(dir_)
        except FileNotFoundError:
            pass
    if not os.path.exists(dir_):
        os.mkdir(dir_)


def validate_directory(dir_, file_ext=None):
    if not isinstance(dir_, str):
        raise TypeError(f"Directory must be a string, got '{type(dir_)}'")
    glob_files = glob.glob(dir_)
    globbing = True
    for file in glob_files:
        if file_ext in file:
            break
    else:
        globbing = False
    if not globbing and len(os.listdir(dir_)) == 0:
        raise ValueError(
            f"Provided directory '{dir_}' does not exist or does not contain any files"
        )


def check_python_verison():
    version_data = sys.version
    if int(version_data.split()[0].split(".")[1]) >= 9:
        return True
    return False


def validate_plink():
    try:
        run_command(["plink", "-h"])
    except RuntimeError:
        raise RuntimeError(
            "PLINK was not found. Must install PLINK to run sample call rate or Hardy-Weinburg QC steps."
        )


def validate_file(file):
    if not isinstance(file, str):
        raise TypeError(f"File path must be a string, got '{type(file)}'")

    if not os.path.exists(file):
        raise FileNotFoundError(f"File '{file}' does not exist")

    if not os.path.isfile(file):
        raise FileNotFoundError(f"Path '{file}' is not a file")

    if not os.access(file, os.R_OK):
        raise PermissionError(f"File '{file}' is not readable")



# MIT License

# Copyright (c) 2024 Andrew Haddad

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
import os
import re
from functools import lru_cache
from io import StringIO

import numpy as np
import pandas as pd

from .common import *
from .utils import blockgz_compress, create_logger, get_files, run_command

logger = create_logger(__name__, stream=True)

GENES_WITH_AS = {"CYP2C9", "DPYD", "CYP2D6"}


def get_cpic_dip_to_phen():
    dip_to_phen_data = pd.read_table(DIP_TO_PHEN_PATH, comment="#")
    dip_to_phen_data_raw = dip_to_phen_data.drop(columns="activity_score")
    dip_to_phen = {
        "VKORC1": {
            "rs9923231 reference (C)/rs9923231 reference (C)": "Variant Absent",
            "rs9923231 reference (C)/rs9923231 variant (T)": "Variant Present",
            "rs9923231 variant (T)/rs9923231 variant (T)": "Variant Present",
            "Indeterminate/Indeterminate": "Indeterminate",
        },
        "2C_CLUSTER": {
            "rs12777823 reference (G)/rs12777823 reference (G)": "Variant Absent",
            "rs12777823 reference (G)/rs12777823 variant (A)": "Variant Present",
            "rs12777823 variant (A)/rs12777823 variant (A)": "Variant Present",
            "not called": "Indeterminate",
            "Indeterminate/Indeterminate": "Indeterminate",
        },
    }
    for _, (gene, dip, phen) in dip_to_phen_data_raw.iterrows():
        try:
            dip_to_phen[gene].update({dip: phen})
        except KeyError:
            dip_to_phen[gene] = {dip: phen}
    dip_to_activity_data = dip_to_phen_data.loc[
        dip_to_phen_data["gene"].isin(GENES_WITH_AS)
    ].drop(columns="phenotype")
    dip_to_activity_score = {
        "DPYD": {
            "Indeterminate/Indeterminate": "n/a",
        },
        "CYP2C9": {
            "Indeterminate/Indeterminate": "n/a",
        },
        "CYP2D6": {"Indeterminate/Indeterminate": "n/a"},
    }
    for _, (gene, dip, activity) in dip_to_activity_data.iterrows():
        dip_to_activity_score[gene].update({dip: activity})
    return dip_to_phen, dip_to_activity_score


def get_allele_function(gene=None):
    if gene == "CYP2D6":
        file_path = CYP2D6_ALLELE_FXN_PATH
    else:
        file_path = DPYD_ALLELE_FXN_PATH
    with open(file_path) as f:
        allele_function = json.load(f)
    return allele_function


def assign_dpyd_phenotype(activity_score):
    if activity_score == "2.0":
        return "Normal Metabolizer"
    elif activity_score in {"1.5", "1.0"}:
        return "Intermediate Metabolizer"
    elif activity_score in {"0.5", "0.0"}:
        return "Poor Metabolizer"
    else:
        return "Indeterminate"


def assign_dpyd_activity_score(alleles=None, dpyd_allele_function=None):
    # Activity score is based on the two alleles with the lowest activity
    # https://pubmed.ncbi.nlm.nih.gov/29152729/
    activity_scores_alleles = [dpyd_allele_function[allele] for allele in alleles]
    return sum(sorted(activity_scores_alleles)[:2])


def adjust_dpyd_reference(genotype):
    # According to CPIC tables order should be allele/Reference
    if genotype.startswith("Reference"):
        alleles = genotype.split("/")
        return alleles[1] + "/" + alleles[0]
    return genotype


def parse_dpyd(gene_data=None, dpyd_allele_function=None, dip_to_phen=None):
    # Samples can have >2 alleles because the gene is so large and does
    # not exist as a single haplotype
    # https://pubmed.ncbi.nlm.nih.gov/29152729/
    if gene_data["diplotypes"]:
        genotypes = set()
        for diplotype_data in gene_data["diplotypes"]:
            diplotype = diplotype_data["name"]
            alleles = (
                diplotype.replace("[", "")
                .replace("]", "")
                .replace(" + ", "/")
                .split("/")
            )
            genotype = "/".join(sorted(alleles))
            genotype = adjust_dpyd_reference(genotype)
            genotypes.add(genotype)
        if len(genotypes) == 1:
            genotype = next(iter(genotypes))
            activity_score = assign_dpyd_activity_score(
                alleles=genotype.split("/"), dpyd_allele_function=dpyd_allele_function
            )
        else:
            genotype = "Indeterminate/Indeterminate"
            activity_score = "n/a"
    else:
        haplotypes = []
        for haplotype in gene_data["haplotypeMatches"]:
            haplotypes.append(haplotype["name"])
        genotype = "/".join(sorted(haplotypes))
        genotype = adjust_dpyd_reference(genotype)
        activity_score = assign_dpyd_activity_score(
            genotype.split("/"), dpyd_allele_function=dpyd_allele_function
        )
    if genotype.count("/") == 1:
        if genotype not in dip_to_phen["DPYD"]:
            first, second = genotype.split("/")
            genotype = f"{second}/{first}"
    return genotype, activity_score


def parse_2c_cluster(gene_data=None, dip_to_phen=None):
    # this "gene" doesn't exist within the normal data
    cluster_genotype_converter = {
        "G/G": "rs12777823 reference (G)/rs12777823 reference (G)",
        "G/A": "rs12777823 reference (G)/rs12777823 variant (A)",
        "A/A": "rs12777823 variant (A)/rs12777823 variant (A)",
        "./.": "Indeterminate/Indeterminate",
        None: "Indeterminate/Indeterminate",
    }
    genotype_raw = gene_data["variantsOfInterest"][0]["vcfCall"]
    genotype = cluster_genotype_converter[genotype_raw]
    phenotype = dip_to_phen["2C_CLUSTER"][genotype]
    return genotype, phenotype


@lru_cache()
def assign_cyp4f2_phenotype(genotype):
    # Phenotype/guideline is based on the variant that defines the *3
    # allele which is also contained in the new *4 allele
    # https://pubmed.ncbi.nlm.nih.gov/28198005/
    if "*3" in genotype or "*4" in genotype:
        return "Variant Present"
    return "Variant Absent"


def adjust_cyp2d6_calls(cyp2d6_cnv, genotype, cyp2d6_allele_function=None):
    phenotype = "Indeterminate"
    activity_score = "n/a"
    if cyp2d6_cnv == 3:
        allele1, allele2 = genotype.split("/")
        if allele1 == allele2:
            genotype = genotype + "xN"
            genotype_3copies = f"{allele1}/{allele2}x2"
            activity_score = assign_cyp2d6_activity(
                genotype=genotype_3copies, cyp2d6_allele_function=cyp2d6_allele_function
            )
            phenotype = assign_cyp2d6_phenotype(activity_score)
        else:
            genotype1 = f"{allele1}/{allele2}x2"
            genotype2 = f"{allele1}x2/{allele2}"
            genotype = genotype + "xN"
            activity_score1 = assign_cyp2d6_activity(
                genotype=genotype1, cyp2d6_allele_function=cyp2d6_allele_function
            )
            activity_score2 = assign_cyp2d6_activity(
                genotype=genotype2, cyp2d6_allele_function=cyp2d6_allele_function
            )
            phenotype1 = assign_cyp2d6_phenotype(activity_score1)
            phenotype2 = assign_cyp2d6_phenotype(activity_score2)
            if phenotype1 == phenotype2:
                phenotype = phenotype1
            else:
                phenotype = "Indeterminate"
            if activity_score1 == activity_score2:
                activity_score = activity_score1
            else:
                activity_score = f"{activity_score1} or {activity_score2}"
    elif cyp2d6_cnv == "CN_HybridLoss":
        genotype = "Indeterminate/Indeterminate"
    elif cyp2d6_cnv == "CN_HybridGain":
        genotype = genotype + " CN_HybridGain"
    return genotype, phenotype, activity_score


def adjust_cyp2d6_gene_deletion(cyp2d6_cnv, genotype, cyp2d6_allele_function=None):
    if cyp2d6_cnv == 0:
        genotype = "*5/*5"
        activity_score = 0
        phenotype = "Poor Metabolizer"
    elif cyp2d6_cnv == 1:
        allele1, allele2 = genotype.split("/")
        if allele1 == "*1":
            allele2_num = int(allele2.replace("*", ""))
            if allele2_num < 5:
                genotype = f"{allele2}/*5"
            else:
                genotype = f"*5/{allele2}"
            activity_score = assign_cyp2d6_activity(
                genotype=genotype, cyp2d6_allele_function=cyp2d6_allele_function
            )
            phenotype = assign_cyp2d6_phenotype(activity_score)
        else:
            genotype = "Indeterminate/Indeterminate"
            activity_score = "n/a"
            phenotype = "Indeterminate"
    return genotype, phenotype, activity_score


@lru_cache(maxsize=None)
def assign_cyp2d6_activity(genotype=None, cyp2d6_allele_function=None):
    if genotype == "Indeterminate/Indeterminate":
        return "n/a"
    elif genotype is np.nan:
        return np.nan
    total_activity = 0
    na_activity = False
    for allele in genotype.split("/"):
        copies = 1
        if "x" in allele:
            copies = int(re.findall(r"x\d", allele)[0].strip("x"))
        allele = re.findall(r"\*\d+", allele)[0]
        allele_activity = cyp2d6_allele_function[allele]
        try:
            total_activity += float(allele_activity) * copies
        except ValueError:
            na_activity = True
    if na_activity:
        return f">= {total_activity}"
    return total_activity


@lru_cache()
def assign_cyp2d6_phenotype(activity_score):
    # https://pubmed.ncbi.nlm.nih.gov/31647186/
    try:
        if activity_score == "n/a" or ">=" in activity_score:
            return "Indeterminate Metabolizer"
    except TypeError:
        # activity is already a float
        if activity_score == 0:
            return "Poor Metabolizer"
        elif activity_score >= 0.25 and activity_score <= 1:
            return "Intermediate Metabolizer"
        elif activity_score >= 1.25 and activity_score <= 2.25:
            return "Normal Metabolizer"
        elif activity_score > 2.25:
            return "Ultrarapid Metabolizer"


def determine_cyp2d6_cnv(file):
    file_gz = file + ".gz"
    file_gz = blockgz_compress(input_file=file, output_file=file_gz)
    try:
        cnv_data = run_command(
            f"bcftools view {file_gz} -H --regions-file {CYP2D6_CNV_POSITIONS_PATH} -Ov",
            capture_output=True,
            shell=True,
        )
        cnvs = (
            pd.read_table(
                StringIO(cnv_data), usecols=[4, 9], header=None, names=["ALT", "sample"]
            )
            .loc[lambda x: x["ALT"] != "C"]
            .iloc[:, -1]
            .astype(np.int64)
            .to_list()
        )
        if len(set(cnvs)) == 1:
            return cnvs[0]
        if 1 in cnvs or 0 in cnvs:
            return "CN_HybridLoss"
        else:
            return "CN_HybridGain"
    finally:
        os.remove(file_gz)


def create_frame(data):
    df = pd.DataFrame(
        data,
        columns=[
            "person_id",
            "gene",
            "genotype",
            "phenotype",
            "activity_score",
            "possible_genotype",
            "possible_phenotype",
        ],
    )
    return df


def get_genotype_phenotype(call=None, gene=None, dip_to_phen=None):
    # May need to adjust order of alleles which changes genotype

    # Weird comma for certain G6PD alleles
    call = call.replace("‚", ",")
    try:
        phenotype = dip_to_phen[gene][call]
    except KeyError:
        genotype_new = call.split("/")
        genotype_new = f"{genotype_new[1]}/{genotype_new[0]}"
        try:
            phenotype = dip_to_phen[gene][genotype_new]
            genotype = genotype_new
        except KeyError:
            genotype = call
            phenotype = "Indeterminate"
    else:
        genotype = call
    return genotype, phenotype


def _parse_pharmcat_output(
    file=None,
    vcf_dir=None,
    xy_ids=None,
    dip_to_phen=None,
    dip_to_activity_score=None,
    dpyd_allele_function=None,
    cyp2d6_allele_function=None,
):
    compiled_data = []
    with open(file) as f:
        sample_id = os.path.basename(file).split(".")[1]
        json_data = json.load(f)
        for gene_data in json_data["results"]:
            gene = gene_data["gene"]
            activity_score = np.nan
            if gene in {"CFTR", "CYP3A4", "IFNL3", "F5"}:
                # Not doing these genes
                continue
            elif gene == "CYP2C9":
                # CYP2C Cluster parsing must occur first otherwise genotype and phenotype are overwritten
                genotype, phenotype = parse_2c_cluster(
                    gene_data=gene_data, dip_to_phen=dip_to_phen
                )
                compiled_data.append(
                    [
                        sample_id,
                        "2C_CLUSTER",
                        genotype,
                        phenotype,
                        np.nan,
                        np.nan,
                        np.nan,
                    ]
                )
            if gene == "DPYD":
                genotype, activity_score = parse_dpyd(
                    gene_data=gene_data,
                    dpyd_allele_function=dpyd_allele_function,
                    dip_to_phen=dip_to_phen,
                )
                phenotype = assign_dpyd_phenotype(str(activity_score))
                possible_genotype, possible_phenotype = np.nan, np.nan
            else:
                genotypes = []
                phenotypes = []
                if xy_ids is not None:
                    if (
                        len(gene_data["diplotypes"]) == 0
                        and sample_id in xy_ids
                        and gene == "G6PD"
                    ):
                        # Assigns single Indeterminate allele for samples that are not called
                        genotypes.append("Indeterminate")
                        phenotypes.append("Indeterminate")
                for call_data in gene_data["diplotypes"]:
                    call = call_data["name"]
                    if gene == "G6PD":
                        if xy_ids is not None and sample_id in xy_ids:
                            call_split = call.split("/")
                            if call_split[0] != call_split[1]:
                                # Scenarios where a variant is not called and therefore a possible
                                # call is made that is not possible
                                # i.e A/A, A/Sierra Leone, Sierra Leone/Sierra Leone
                                # A/Sierra Leone is not possible for XY sample
                                continue
                            call = call_split[0]
                    genotype, phenotype = get_genotype_phenotype(
                        call=call, gene=gene, dip_to_phen=dip_to_phen
                    )
                    genotypes.append(genotype)
                    phenotypes.append(phenotype)
                if len(genotypes) == 1:
                    genotype = genotypes[0]
                    possible_genotype, possible_phenotype = np.nan, np.nan
                    phenotype = phenotypes[0]
                else:
                    if gene == "G6PD" and xy_ids is not None and sample_id in xy_ids:
                        genotype = "Indeterminate"
                    else:
                        genotype = "Indeterminate/Indeterminate"
                    possible_genotype = (
                        ";".join(genotypes) if len(genotypes) != 0 else np.nan
                    )
                    phenotype = "Indeterminate"
                    possible_phenotype = np.nan
                    if len(set(phenotypes)) == 1:
                        possible_phenotype = phenotypes[0]
                if gene in {"CYP2C9", "CYP2D6"}:
                    try:
                        activity_score = dip_to_activity_score[gene][genotype]
                    except KeyError:
                        activity_score = "n/a"
                if gene == "CYP4F2":
                    phenotype = assign_cyp4f2_phenotype(genotype)
                if gene == "CYP2D6":
                    if vcf_dir is None:
                        cyp2d6_cnv = 2
                    else:
                        cyp2d6_file = os.path.join(
                            vcf_dir.replace("*.vcf", ""),
                            f"No_SULT1A1_{sample_id}.CEL.vcf",
                        )

                        cyp2d6_cnv = determine_cyp2d6_cnv(file=cyp2d6_file)
                    if (
                        cyp2d6_cnv not in {0, 1, 2}
                        and genotype != "Indeterminate/Indeterminate"
                    ):
                        genotype, phenotype, activity_score = adjust_cyp2d6_calls(
                            cyp2d6_cnv=cyp2d6_cnv,
                            genotype=genotype,
                            cyp2d6_allele_function=cyp2d6_allele_function,
                        )
                    elif cyp2d6_cnv in {0, 1}:
                        genotype, phenotype, activity_score = (
                            adjust_cyp2d6_gene_deletion(
                                cyp2d6_cnv=cyp2d6_cnv,
                                genotype=genotype,
                                cyp2d6_allele_function=cyp2d6_allele_function,
                            )
                        )
            compiled_data.append(
                [
                    sample_id,
                    gene,
                    genotype,
                    phenotype,
                    activity_score,
                    possible_genotype,
                    possible_phenotype,
                ]
            )
    df = create_frame(compiled_data)
    return df


def finalize_results(results):
    call_data_pharmcat = pd.concat(results, axis=0, ignore_index=True)
    call_data_pharmcat = call_data_pharmcat.replace(
        "not called", "Indeterminate/Indeterminate"
    )
    call_data_pharmcat = call_data_pharmcat.replace("N/A", "Indeterminate")
    call_data_pharmcat.loc[:, "phenotype"] = (
        call_data_pharmcat.loc[:, "gene"] + " " + call_data_pharmcat.loc[:, "phenotype"]
    )
    call_data_pharmcat.loc[:, "possible_phenotype"] = (
        call_data_pharmcat.loc[:, "gene"]
        + " "
        + call_data_pharmcat.loc[:, "possible_phenotype"]
    )

    cond = call_data_pharmcat["gene"].isin(GENES_WITH_AS)
    call_data_pharmcat.loc[cond, "activity_score"] = (
        call_data_pharmcat.loc[cond, "gene"]
        + " Activity Score "
        + call_data_pharmcat.loc[cond, "activity_score"].astype(str)
    )
    return call_data_pharmcat


def write_pharmcat_output(call_data_pharmcat=None, dir_=None, output_dir=None):
    fname = "call_data_pharmcat.csv"
    if output_dir:
        output_path = os.path.join(output_dir, fname)
    else:
        if "*" in dir_:
            output_path = os.path.join(os.path.dirname(dir_), fname)
        else:
            output_path = os.path.join(".", fname)
    call_data_pharmcat.to_csv(output_path, index=False)


def check_sex_ids(json_files, sex_data):
    if len(json_files) != sex_data.shape[0]:
        raise ValueError(
            f"The number of json files {len(json_files)} does not match the number of samples in the sex ids file {sex_data.shape[0]}"
        )
    json_sample_ids = set()
    for json_file in json_files:
        sample_id = json_file.split(".")[1]
        json_sample_ids.add(sample_id)

    sex_ids = set(sex_data[0].astype(str))

    diff = json_sample_ids.difference(sex_ids)
    if len(diff) != 0:
        raise ValueError(f"Found {diff} sample ids in JSONs not in sex_ids file")
    diff = sex_ids.difference(json_sample_ids)
    if len(diff) != 0:
        raise ValueError(
            f"Found {diff} sample ids in sex_ids with no matching JSON file"
        )


def parse_pharmcat_output(
    pharmcat_dir=None, vcf_dir=None, sex_ids=None, output_dir=None
):
    if sex_ids is None:
        logger.warn(
            "No sex info was provided. Without this, G6PD genotypes and phenotypes can be innacurate."
        )
        xy_ids = None
    else:
        sex_data = pd.read_table(sex_ids, header=None)
    if vcf_dir is None:
        logger.warn(
            "No VCF directory was provided. Without this, CYP2D6 copy number cannot be determined including *5."
        )
    files = [file for file in get_files(dir_=pharmcat_dir) if ".json" in file]
    if sex_ids:
        check_sex_ids(json_files=files, sex_data=sex_data)
        xy_ids = set(sex_data.loc[sex_data[1] == "M", 0].astype(str))

    dip_to_phen, dip_to_activity_score = get_cpic_dip_to_phen()
    dpyd_allele_function = get_allele_function(gene="DPYD")
    cyp2d6_allele_function = get_allele_function(gene="CYP2D6")
    parsed_results = []
    for file in files:
        try:
            result = _parse_pharmcat_output(
                file=file,
                vcf_dir=vcf_dir,
                xy_ids=xy_ids,
                dip_to_phen=dip_to_phen,
                dip_to_activity_score=dip_to_activity_score,
                dpyd_allele_function=dpyd_allele_function,
                cyp2d6_allele_function=cyp2d6_allele_function,
            )
            parsed_results.append(result)
            logger.info(f"Parsed results for {file}")
        except Exception as e:
            logger.exception(f"Failed to parse results for {file}")
            raise e
    call_data_pharmcat = finalize_results(parsed_results)
    write_pharmcat_output(
        call_data_pharmcat=call_data_pharmcat, dir_=pharmcat_dir, output_dir=output_dir
    )



# MIT License

# Copyright (c) 2024 Andrew Haddad

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import argparse
import os
import sys

from .haplotype import call_haplotypes
from .install import install
from .parser import parse_pharmcat_output
from .utils import (
    validate_directory,
    validate_environment,
    validate_plink,
    validate_file,
)
from .vcf import standardize_vcfs
from .common import MANIFEST_HEADER


def call_install(args):
    if "linux" not in sys.platform.lower():
        raise OSError("Only linux systems are supported at this time.")
    install(overwrite=args.overwrite)


def call_haplotype(args):
    validate_environment()
    validate_directory(args.files, file_ext=".vcf")
    if args.vcf_standardized:
        validate_file(args.vcf_standardized)
    if args.hwe or args.sample_call_rate:
        validate_plink()
    call_haplotypes(
        dir_=args.files,
        vcf_standardized=args.vcf_standardized,
        hwe=args.hwe,
        variant_call_rate=args.variant_call_rate,
        sample_call_rate=args.sample_call_rate,
        output_dir=args.output,
        debug=args.debug,
    )


def call_parser(args):
    validate_directory(args.files, file_ext=".json")
    if args.sex_ids:
        validate_file(args.sex_ids)
    if args.vcfs:
        validate_directory(args.vcfs, file_ext=".vcf")
    parse_pharmcat_output(
        pharmcat_dir=args.files,
        vcf_dir=args.vcfs,
        sex_ids=args.sex_ids,
        output_dir=args.output,
    )


def call_standardize_vcfs(args):
    validate_directory(args.files, file_ext=".vcf")
    standardize_vcfs(files=args.files, output_dir=args.output)


def main():
    parser = argparse.ArgumentParser(
        description="Runs PharmCAT pipeline using Pharmacoscan input data",
    )
    subparsers = parser.add_subparsers(required=True)

    # Install
    subparser_install = subparsers.add_parser(
        "install", help="Install PharmCAT pipeline"
    )
    subparser_install.set_defaults(func=call_install)
    subparser_install.add_argument(
        "--overwrite",
        help="Overwrite existing pharmcat installation",
        action="store_true",
        default=False,
    )

    # standardize_vcf
    subparser_standardize = subparsers.add_parser(
        "standardize", help="Standardize VCFs to multisample GRCh38 and v4.2 spec"
    )
    subparser_standardize.set_defaults(func=call_standardize_vcfs)

    subparser_standardize.add_argument(
        "--files",
        help="Directory of VCF file inputs. Supports with and without glob patterns.",
        type=str,
        default=None,
    )
    subparser_standardize.add_argument(
        "--output",
        help="Output directory. Default is results within the working directory.",
        type=str,
        default=os.getcwd(),
    )

    # Haplotype
    subparser_haplotype = subparsers.add_parser(
        "haplotype", help="Call haplotypes using PharmCAT"
    )
    subparser_haplotype.set_defaults(func=call_haplotype)
    group = subparser_haplotype.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--files",
        help="Directory of VCF file inputs. Supports with and without glob patterns.",
        type=str,
    )
    group.add_argument("--vcf", help="Standardized VCF input", type=str, default=None)

    subparser_haplotype.add_argument(
        "--hwe",
        type=float,
        help="Cutoff for Hardy-Weinburg Equilibrium. PLINK must be installed and callable to run this QC method.",
    )
    subparser_haplotype.add_argument(
        "--variant_call_rate", type=float, help="Variant call rate cutoff"
    )
    subparser_haplotype.add_argument(
        "--sample_call_rate", type=float, help="Sample call rate cutoff"
    )
    subparser_haplotype.add_argument(
        "--debug",
        action="store_true",
        default=False,
        help="Run haplotyping in debug mode.",
    )
    subparser_haplotype.add_argument(
        "--output",
        help="Output directory. Default is results within the working directory.",
        type=str,
        default=os.path.join(os.getcwd(), "results"),
    )

    # Parser
    subparser_pharmcat_parser = subparsers.add_parser(
        "parse", help="Parse PharmCAT output"
    )
    subparser_pharmcat_parser.set_defaults(func=call_parser)
    subparser_pharmcat_parser.add_argument(
        "--files",
        help="Directory of PharmCAT JSON files. Supports with and without glob patterns.",
        required=True,
        type=str,
    )
    subparser_pharmcat_parser.add_argument(
        "--sex_ids",
        help="Path to file to with genetically imputed sex for each sample. Required to accurately call G6PD",
        type=str,
    )
    subparser_pharmcat_parser.add_argument(
        "--vcfs",
        help="Path to VCFs. Required for accurate CYP2D6 haplotyping.",
        type=str,
    )
    subparser_pharmcat_parser.add_argument(
        "--output",
        help="Directory to save pharmcat parsed output",
        type=str,
        default="results",
    )
    args = parser.parse_args()
    print(MANIFEST_HEADER)
    args.func(args)


if __name__ == "__main__":
    main()



# MIT License

# Copyright (c) 2024 Andrew Haddad

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import shutil
import sys
import traceback
from tempfile import TemporaryDirectory

from .common import *
from .utils import curl, run_command, unpack_tar, check_python_verison, create_logger


logger = create_logger(__name__, stream=True)


def install_java(tempdir):
    curl(tempdir, JAVA_URL)
    unpack_tar(file=os.path.join(tempdir, os.path.basename(JAVA_URL)), dir_=BIN_DIR)


def install_python(tempdir):
    curl(tempdir, PYTHON_URL)
    unpack_tar(file=os.path.join(tempdir, os.path.basename(PYTHON_URL)), dir_=tempdir)
    python_dir = os.path.basename(PYTHON_URL).replace(".tgz", "")
    python_dir_current = os.path.join(tempdir, python_dir)
    run_command(
        f"cd {python_dir_current} && ./configure --prefix={os.path.join(BIN_DIR, 'python')}",
        shell=True,
    )
    run_command(f"cd {python_dir_current} && make", shell=True)
    run_command(f"cd {python_dir_current} && make altinstall", shell=True)


def install_pharmcat():
    dir_ = os.path.join(BIN_DIR, "preprocessor")
    curl(dir_, PHARMCAT_JAR_URL)
    os.rename(
        os.path.join(dir_, os.path.basename(PHARMCAT_JAR_URL)),
        os.path.join(dir_, "pharmcat.jar"),
    )


def download_reference_genome(tempdir):
    curl(dir_=tempdir, url=REFERENCE_GENOME_URL)
    unpack_tar(
        file=os.path.join(tempdir, os.path.basename(REFERENCE_GENOME_URL)),
        dir_=os.path.join(BIN_DIR, "preprocessor"),
    )


def install_preprocessor(tempdir):
    # TODO resolve pipenv/dependency issue on MAC; works on linux
    # version = re.findall(r"\d\.\d+", os.path.basename(PYTHON_URL))[0]
    # python_path = os.path.join(BIN_DIR, "python", "bin", f"python{version}")
    python_path = "python3"
    run_command([python_path, "-m", "pip", "install", "-U", "pipenv", "--user"])
    curl(tempdir, PHARMCAT_PREPROCESSOR_URL)
    unpack_tar(
        file=os.path.join(tempdir, os.path.basename(PHARMCAT_PREPROCESSOR_URL)),
        dir_=BIN_DIR,
    )
    run_command(
        f"cd {os.path.join(BIN_DIR, 'preprocessor')} && {python_path} -m pipenv install -r requirements.txt",
        shell=True,
    )


def set_env_vars():
    bash_profile = os.path.join(os.path.expanduser("~"), ".bash_profile")
    bash_profile_temp = os.path.join(os.path.expanduser("~"), ".bash_profile_temp")
    with open(bash_profile) as f_in:
        with open(bash_profile_temp, "w") as f_out:
            for line in f_in:
                if "JAVA_HOME" not in line:
                    f_out.write(line)
            for file in os.listdir(BIN_DIR):
                if "jdk" in file:
                    f_out.write(f'export JAVA_HOME="{os.path.join(BIN_DIR, file)}"\n')
    shutil.copy2(bash_profile_temp, bash_profile)
    os.remove(bash_profile_temp)


def validate_python_version(tempdir):
    if not check_python_verison():
        try:
            logger.info(
                "Incorrect version of Python was found. Attempting to install Python..."
            )
            install_python(tempdir)
        except Exception:
            try:
                shutil.rmtree(os.path.join(BIN_DIR, "python"))
            except FileNotFoundError:
                pass
            finally:
                msg = f"Requires Python v3.9+, found v{sys.version.split()[0]}. Attempt to install new version of Python was unsuccesful"
                logger.error(msg)
                raise RuntimeError(msg)
        else:
            logger.info("Successfully installed Python.")


def install(overwrite=False):
    try:
        os.mkdir(BIN_DIR)
    except FileExistsError:
        if overwrite:
            logger.info("Removing old version of PharmCAT")
            shutil.rmtree(BIN_DIR)
            logger.info("Succesfully removed old version of PharmCAT")
            os.mkdir(BIN_DIR)
        else:
            msg = "Previous installation of PharmCAT found. To install a new version, must supply --overwrite"
            logger.error(msg)
            raise FileExistsError(msg)
    try:
        with TemporaryDirectory() as tempdir:
            logger.info("***Installing PharmCAT and all required dependencies***")

            logger.info("Installing java...")
            install_java(tempdir)
            logger.info("Successfully installed java.")

            validate_python_version(tempdir)

            logger.info("Installing Preprocessor and PharmCAT pipeline...")
            install_preprocessor(tempdir)
            logger.info("Successfully installed preprocessor and PharmCAT pipeline.")

            logger.info("Downloading reference genome...")
            download_reference_genome(tempdir)
            logger.info("Succesfully downloaded reference genome.")

            logger.info("Installing PharmCAT...")
            install_pharmcat()
            logger.info("Succesfully installed PharmCAT.")

            set_env_vars()
            logger.info("Succesfully set environment variables.")
            logger.info("***Installation completed succesfully.***")
            logger.info(
                "It is recommended to restart the environment/terminal for installation to finalize."
            )
    except Exception:
        logger.exception("Installation failed")



# MIT License

# Copyright (c) 2024 Andrew Haddad

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import glob
import os
import shutil
from datetime import datetime
from tempfile import TemporaryDirectory

from .common import *
from .qc_metrics import QCMetrics
from .utils import (
    check_python_verison,
    create_logger,
    get_files,
    maybe_create_dir,
    run_command,
)
from .vcf import filter_vcf_for_variants, standardize_vcfs

logger = create_logger(__name__, stream=True)


def _copy_outputs(dir_, glob_pattern, tempdir):
    maybe_create_dir(dir_)
    for file in glob.glob(os.path.join(tempdir, glob_pattern)):
        shutil.copy2(file, dir_)


def move_outputs(vcf_input=None, tempdir=None, output_dir=None):
    # VCF used in PharmCAT
    shutil.copy2(vcf_input, output_dir)
    shutil.copy2(vcf_input + ".tbi", output_dir)

    # Log file
    shutil.copy2(os.path.join(os.getcwd(), "pharmcat_runner.log"), output_dir)

    data_to_copy = {
        "pharmcat_output": "*.json",  # PharmCAT jsons
        "pharmcat_warnings": "*warnings*",  # PharmCAT warnings
        "qc_metrics": "*.tsv",  # QC metrics
    }
    for dir_, glob_pattern in data_to_copy.items():
        dir_ = os.path.join(output_dir, dir_)
        _copy_outputs(dir_=dir_, glob_pattern=glob_pattern, tempdir=tempdir)


def get_python_path():
    if not check_python_verison():
        return os.path.join(BIN_DIR, "python", "bin", "python3.9")
    return "python3"


def run_pharmcat(vcf_input, tempdir):
    preprocessor_path = os.path.join(BIN_DIR, "preprocessor")
    python_path = get_python_path()
    pharmcat_pipeline_path = os.path.join(preprocessor_path, "pharmcat_pipeline")
    num_threads = os.cpu_count() - 1 if os.cpu_count() != 1 else 1
    run_command(
        f"cd {preprocessor_path} && {python_path} -m pipenv run python {pharmcat_pipeline_path} {vcf_input} --missing-to-ref -o {tempdir} -matcher --research-mode cyp2d6 -cp {num_threads}",
        shell=True,
    )


def get_pharmcat_positions(tempdir=None):
    # Multiallelic variants mess up indels
    # Need to split these up so that they are captured correctly

    pharmcat_variants_orig_file = os.path.join(
        BIN_DIR, "preprocessor", "pharmcat_positions.vcf"
    )
    pharmcat_variants_split_file = os.path.join(tempdir, "pharmcat_positions_split.vcf")
    pharmcat_positions_file = os.path.join(tempdir, "pharmcat_positions.tsv.gz")

    # Splits multiallelic and left normalizes pharmcat variants
    run_command(
        f"bcftools norm -m -both -c e -f {REFERENCE_PATH} {pharmcat_variants_orig_file} > {pharmcat_variants_split_file} && bcftools query -f'%CHROM\t%POS\t%REF,%ALT\n' {pharmcat_variants_split_file} | bgzip -c > {pharmcat_positions_file} && tabix -s1 -b2 -e2 {pharmcat_positions_file}",
        shell=True,
    )

    return pharmcat_positions_file


def write_manifest(
    output_dir=None,
    num_files=None,
    vcf_input=None,
    variant_call_rate=None,
    sample_call_rate=None,
    hwe=None,
):
    for file in os.listdir(BIN_DIR):
        if "jdk" in file:
            jdk_version = file.replace("jdk-", "")
    user = os.path.basename(os.path.expanduser("~"))
    today = datetime.today().strftime("%B %d, %Y")
    bcftools_version = (
        run_command(["bcftools", "--version"], capture_output=True)
        .split("\n")[0]
        .split(" ")[-1]
        .strip()
    )
    num_samples = run_command(
        f"bcftools query -l {vcf_input} | wc -l", capture_output=True, shell=True
    ).strip()
    plink_version = run_command(["plink", "--version"], capture_output=True).split()[1]
    with open(os.path.join(BIN_DIR, "preprocessor", "preprocessor", "common.py")) as f:
        for line in f:
            if "PHARMCAT_VERSION" in line:
                pharmcat_version = line.split("= ")[-1].strip().strip("'")
    with open(os.path.join(output_dir, "manifest.txt"), "w") as f:
        f.write(MANIFEST_HEADER + "\n")
        f.write(f"User: {user}\n")
        f.write(f"Date: {today}\n")
        f.write(f"PharmCAT version: {pharmcat_version}\n")
        f.write(f"Python version: {sys.version.split()[0]}\n")
        f.write(f"JDK version: {jdk_version}\n")
        f.write(f"Bcftools version: {bcftools_version}\n")
        f.write(f"PLINK version: {plink_version}\n")
        f.write(f"Number of input VCF files: {num_files}\n")
        f.write(f"Number of samples: {num_samples}\n")
        f.write(f"Variant call rate cutoff: {variant_call_rate}\n")
        f.write(f"Sample call rate cutoff: {sample_call_rate}\n")
        f.write(f"Hardy Weinburg Equilibrium cutoff: {hwe}\n")
        f.write(f"VCF input to PharmCAT: {os.path.basename(vcf_input)}\n")


def _get_files(dir_):
    files = get_files(dir_)
    clean_files = []
    for file in files:
        if file.endswith(".vcf") or file.endswith(".vcf.gz"):
            clean_files.append(file)
    return clean_files


def move_all_files(files):
    debug_dir = os.path.join(os.getcwd(), "debug")
    maybe_create_dir(debug_dir, strict=True)
    for file in files:
        shutil.copy2(file, debug_dir)


def check_warnings(dir_):
    num_warnings = len(os.listdir(dir_))
    if num_warnings != 0:
        logger.warn(
            f"PharmCAT returned {num_warnings} warnings when running. Review warnings at {dir_}"
        )


def run_qc(
    file=None, dir_=None, hwe=None, variant_call_rate=None, sample_call_rate=None
):
    qc_metrics = QCMetrics(
        file=file,
        dir_=dir_,
        hwe=hwe,
        variant_call_rate=variant_call_rate,
        sample_call_rate=sample_call_rate,
    )
    if variant_call_rate or hwe:
        qc_metrics.variant_qc()
    if sample_call_rate:
        qc_metrics.sample_qc()
    return qc_metrics


def call_haplotypes(
    dir_=None,
    vcf_standardized=None,
    hwe=None,
    variant_call_rate=None,
    sample_call_rate=None,
    output_dir=None,
    debug=False,
):
    with TemporaryDirectory() as tempdir:
        try:
            pharmcat_positions_file = get_pharmcat_positions(tempdir=tempdir)

            logger.info("Prepping files for PharmCAT...")
            if dir_:
                files = _get_files(dir_)
                vcf_standardized = standardize_vcfs(files=files, output_dir=tempdir)
            pharmcat_variants_only_file = filter_vcf_for_variants(
                file=vcf_standardized,
                dir_=tempdir,
                positions_file=pharmcat_positions_file,
                exclusions='POS==42126599 || ID=="AX-165885131_chrX_154532990_CGGT_C"',
            )

            if any([hwe, variant_call_rate, sample_call_rate]):
                logger.info("Running pre-PharmCAT QC...")
                qc_metrics = run_qc(
                    file=pharmcat_variants_only_file,
                    dir_=tempdir,
                    hwe=hwe,
                    variant_call_rate=variant_call_rate,
                    sample_call_rate=sample_call_rate,
                )
            try:
                vcf_input = qc_metrics.output_file
            except NameError:
                vcf_input = pharmcat_variants_only_file
            logger.info("Running PharmCAT...")
            if output_dir is None:
                output_dir = os.path.join(os.getcwd(), "results")
            else:
                if not output_dir.startswith("/"):
                    output_dir = os.path.join(os.getcwd(), output_dir)
            maybe_create_dir(output_dir)
            run_pharmcat(vcf_input=vcf_input, tempdir=tempdir)
            logger.info("Succesfully ran PharmCAT.")
            move_outputs(vcf_input=vcf_input, tempdir=tempdir, output_dir=output_dir)
            check_warnings(dir_=os.path.join(output_dir, "pharmcat_warnings"))
            write_manifest(
                output_dir=output_dir,
                num_files=len(files),
                vcf_input=vcf_input,
                variant_call_rate=variant_call_rate,
                sample_call_rate=sample_call_rate,
                hwe=hwe,
            )
        finally:
            if debug:
                debug_files = glob.glob(os.path.join(tempdir, "*"))
                move_all_files(debug_files)



# MIT License

# Copyright (c) 2024 Andrew Haddad

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import filecmp
import os
import unittest
from shutil import rmtree

from pharmcat_runner import haplotype

PATH = os.path.dirname(os.path.abspath(__file__))


class TestFiles(unittest.TestCase):
    def setUp(self):
        haplotype.call_haplotypes(
            dir_=os.path.join(PATH, "test_files, *.vcf"),
            output_dir=os.path.join(PATH, "test_files", "test_output"),
        )

    def tearDown(self):
        rmtree(os.path.join(PATH, "test_files", "test_output"))

    def test_pipeline(self):
        variant_call_rate = 0.95
        sample_call_rate = 0.95
        hwe = 0.001
        f1 = os.path.join(PATH, "test_files", "something.vcf.gz")
        f2 = os.path.join(PATH, "test_files", "test_output", "something.vcf.gz")
        self.assertTrue(filecmp.cmp(f1, f2, shallow=False))

    def test_pharmcat(self):
        f1 = os.path.join(PATH, "test_files", "something.vcf.gz")
        f2 = os.path.join(PATH, "test_files", "test_output", "something.vcf.gz")
        self.assertTrue(filecmp.cmp(f1, f2, shallow=False))



# MIT License

# Copyright (c) 2024 Andrew Haddad

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import sys

PHARMCAT_JAR_URL = "https://github.com/PharmGKB/PharmCAT/releases/download/v2.11.0/pharmcat-2.11.0-all.jar"
PHARMCAT_PREPROCESSOR_URL = "https://github.com/PharmGKB/PharmCAT/releases/download/v2.11.0/pharmcat-preprocessor-2.11.0.tar.gz"
JAVA_URL = "https://github.com/adoptium/temurin17-binaries/releases/download/jdk-17.0.4.1%2B1/OpenJDK17U-jdk_x64_linux_hotspot_17.0.4.1_1.tar.gz"
PYTHON_URL = "https://www.python.org/ftp/python/3.9.15/Python-3.9.15.tgz"
REFERENCE_GENOME_URL = (
    "https://zenodo.org/record/7288118/files/GRCh38_reference_fasta.tar"
)
BIN_DIR = os.path.join(os.path.expanduser("~"), "pharmcat_bin")
HTS_LIB_REQUIRED_VERSION = 1.18

lib_path = os.path.join(os.path.dirname(sys.modules[__name__].__file__), "lib")

REFERENCE_PATH = os.path.join(BIN_DIR, "preprocessor", "reference.fna.bgz")
CHR_CONV_PATH = os.path.join(lib_path, "chr_name_conv.txt")
DIP_TO_PHEN_PATH = os.path.join(lib_path, "dip_to_phen.tsv")
DPYD_ALLELE_FXN_PATH = os.path.join(lib_path, "dpyd_allele_function.json")
CYP2D6_ALLELE_FXN_PATH = os.path.join(
    lib_path,
    "cyp2d6_allele_function.json",
)
CYP2D6_CNV_POSITIONS_PATH = os.path.join(lib_path, "cyp2d6_cnv_positions.tsv")

# https://www.asciiart.eu/text-to-ascii-art
# Big with PlusBox
MANIFEST_HEADER = """
+--------------------------------------------------------+
| _____  _                           _____       _______ |
||  __ \| |                         / ____|   /\|__   __||
|| |__) | |__   __ _ _ __ _ __ ___ | |       /  \  | |   |
||  ___/| '_ \ / _` | '__| '_ ` _ \| |      / /\ \ | |   |
|| |    | | | | (_| | |  | | | | | | |____ / ____ \| |   |
||_|___ |_| |_|\__,_|_|  |_| |_| |_|\_____/_/    \_\_|   |
||  __ \                                                 |
|| |__) |   _ _ __  _ __   ___ _ __                      |
||  _  / | | | '_ \| '_ \ / _ \ '__|                     |
|| | \ \ |_| | | | | | | |  __/ |                        |
||_|  \_\__,_|_| |_|_| |_|\___|_|                        |
+--------------------------------------------------------+""".strip()

LOGGING_FORMAT = "[%(levelname)s]:%(asctime)s:%(message)s"



# MIT License

# Copyright (c) 2024 Andrew Haddad

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import gzip
import shutil
from functools import wraps

import pandas as pd

from .common import *
from .utils import blockgz_compress, create_logger, log, run_command

logger = create_logger(__name__)


def index_file(func):
    @wraps(func)
    def inner(*args, **kwargs):
        file = func(*args, **kwargs)
        run_command(["tabix", "-p", "vcf", file])
        return file

    return inner


@index_file
@log(logger=logger)
def annotate_id(file):
    # Need to annotate the ID field with something more informative because
    # of QC steps with PLINK. If not, can get rows that have the same affy snp
    # id and no clear way to differentiate them
    output_file = file.replace(".vcf", "_annotateid.vcf")
    run_command(
        [
            "bcftools",
            "annotate",
            "-I",
            "%ID\_%CHROM\_%POS\_%REF\_%ALT",
            "-Oz",
            "-o",
            output_file,
            file,
        ]
    )
    return output_file


@index_file
@log(logger=logger)
def sort_vcf(file, output):
    # Only one where .gz needs to be adjusted because it's the first step
    if ".gz" in output:
        sorted_file = output.replace(".vcf", "_sorted.vcf")
    else:
        sorted_file = output.replace(".vcf", "_sorted.vcf.gz")
    run_command(["bcftools", "sort", "-Oz", "-o", sorted_file, file])
    return sorted_file


@index_file
@log(logger=logger)
def fix_ugt1a1_indels(file):
    # UGT1A1 ref and alt alleles are not correct for the indels
    # Confirmed with email from Carsten 03/28/2024

    # UGT1A1 hom ref and hom alt are flipped
    # Confirmed with email from Carsten 05/08/2024

    # Both will be fixed in r10 release of Pscan
    output_file = file.replace(".vcf", "_ugt1a1fixed.vcf")
    output_file_nogz = output_file.replace(".gz", "")

    # Fix the indel
    new = "TAT\tT,TATAT"
    old = "TATAT\tT,TAT"
    file_data = []
    with gzip.open(file, "rt") as f_in:
        for line in f_in:
            if "233760235" in line:
                if old not in line:
                    return file
                line = line.replace(old, new)
                line_data_raw = line.split("\t")
                line_data_new = []
                for token in line_data_raw:
                    if token == "0/0":
                        line_data_new.append("2/2")
                    elif token == "2/2":
                        line_data_new.append("0/0")
                    elif token == "1/2":
                        line_data_new.append("0/1")
                    else:
                        line_data_new.append(token)
                line = "\t".join(line_data_new)
            file_data.append(line)
    with open(output_file_nogz, "w") as f_out_nogz:
        for line in file_data:
            f_out_nogz.write(line)

    # Compress the output
    # Can't use python gzip to do block gzip compression
    output_file = blockgz_compress(input_file=output_file_nogz, output_file=output_file)

    return output_file


@index_file
@log(logger=logger)
def fix_chromosome_labels(file):
    fixed_file = file.replace(".vcf", "_chrfixed.vcf")
    run_command(
        [
            "bcftools",
            "annotate",
            "--rename-chr",
            CHR_CONV_PATH,
            "-Oz",
            "-o",
            fixed_file,
            file,
        ]
    )
    return fixed_file


@index_file
@log(logger=logger)
def split_multiallelic(file):
    # Needed because this can throw pharmcat off
    split_file = file.replace(".vcf", "_split.vcf")
    run_command(
        [
            "bcftools",
            "norm",
            "-m-any",
            "-c",
            "e",
            "-f",
            REFERENCE_PATH,
            "-Oz",
            "-o",
            split_file,
            file,
        ]
    )
    return split_file


@index_file
@log(logger=logger)
def filter_vcf_for_variants(file, dir_=None, positions_file=None, exclusions=None):
    # Selects only pharmcat variants
    # Including non pharmcat variants can cause issues with calling
    cleaned_file = os.path.join(dir_, "pharmcat_variants_only.vcf.gz")

    # Can't use "--regions-overlap variant" because indels don't get captured correctly
    cmd = ["bcftools", "view", "--regions-file", positions_file]
    rest = ["-Oz", "-o", cleaned_file, file]
    if exclusions:
        cmd = cmd + ["-e", exclusions]
    cmd = cmd + rest

    run_command(cmd)
    return cleaned_file


@index_file
@log(logger=logger)
def merge_vcfs(files, dir_):
    files_to_merge = os.path.join(dir_, "files_to_merge.txt")
    with open(files_to_merge, "w") as f:
        for file in files:
            f.write(file + "\n")

    merged_file = os.path.join(dir_, "merged.vcf.gz")
    run_command(
        [
            "bcftools",
            "merge",
            "--file-list",
            files_to_merge,
            "-i",
            "-",
            "-m",
            "none",
            "-Oz",
            "-o",
            merged_file,
        ]
    )
    return merged_file


@index_file
@log(logger=logger)
def filter_certain_chromosomes(file, reg=True):
    # Need to split out chromsomes that wouldn't be in reference sequence
    # Includes MT and any alternate contigs
    chrs = pd.read_table(CHR_CONV_PATH, header=None, sep=" ")[1]
    if reg:
        output_file = file.replace(".vcf", "_reg_only.vcf")
        chrs = chrs.loc[lambda x: ~x.str.contains("MT|alt", regex=True)]
    else:
        output_file = file.replace(".vcf", "_nonreg_only.vcf")
        chrs = chrs.loc[lambda x: x.str.contains("MT|alt", regex=True)]

    run_command(
        ["bcftools", "view", "-r", ",".join(chrs), "-Oz", "-o", output_file, file]
    )
    return output_file


@index_file
@log(logger=logger)
def concat_vcfs(dir_, *files):
    output_file = os.path.join(dir_, "aligned_file.vcf.gz")
    run_command(["bcftools", "concat", "-a", "-Oz", "-o", output_file] + list(files))
    return output_file


def standardize_vcfs(files, output_dir):
    sorted_files = []
    for file in files:
        sorted_file = sort_vcf(file, os.path.join(output_dir, os.path.basename(file)))
        sorted_files.append(sorted_file)
    merged_file = merge_vcfs(files=sorted_files, dir_=output_dir)
    merged_file_ugt1a1_fixed = fix_ugt1a1_indels(file=merged_file)
    fixed_file = fix_chromosome_labels(merged_file_ugt1a1_fixed)

    reg_chr_only = filter_certain_chromosomes(file=fixed_file, reg=True)
    nonreg_chr_only = filter_certain_chromosomes(file=fixed_file, reg=False)
    split_file = split_multiallelic(file=reg_chr_only)
    alignment_fixed_file = concat_vcfs(output_dir, split_file, nonreg_chr_only)
    alignment_sorted = sort_vcf(
        alignment_fixed_file, os.path.join(output_dir, "alignment.vcf.gz")
    )
    annotate_id_file = annotate_id(file=alignment_sorted)

    standardized_file = os.path.join(
        os.path.dirname(annotate_id_file), "standardized.vcf.gz"
    )
    shutil.copy2(annotate_id_file, standardized_file)
    shutil.copy2(annotate_id_file + ".tbi", standardized_file + ".tbi")

    return standardized_file



# MIT License

# Copyright (c) 2024 Andrew Haddad

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import csv
import os

import pandas as pd

from .utils import log, run_command, create_logger
from .vcf import index_file

logger = create_logger(__name__)


class QCMetrics:
    # Order matters in the QC metric operations
    # Genotype metrics -> variant metrics -> sample metrics
    # Therefore they cannot all be computed at the same time

    # Metrics are first computed using PLINK to generate the required metrics files
    # Then the data is filtered using either bcftools or PLINK
    plink_vcf_flags = [
        "--recode",
        "vcf-iid",  # Uses only sample id for header
        "bgz",
        "--output-chr",  # Outputs chromsomes with chr##
        "chrMT",
        "--real-ref-alleles",  # Needed to code the correct allele as reference
    ]

    attrs_created = ["sc_file", "hwe_file", "vc_file"]

    def __init__(
        self, file, dir_, hwe=None, variant_call_rate=None, sample_call_rate=None
    ):
        self.file = file
        self.dir_ = dir_

        self.variant_call_rate = variant_call_rate
        self.hwe = hwe
        self.sample_call_rate = sample_call_rate

    @staticmethod
    def _parse_metrics_file(file):
        compiled_data = []
        with open(file) as f:
            csv_reader = csv.reader(f, delimiter=" ")
            for line in csv_reader:
                vals = [val for val in line if len(val) != 0]
                compiled_data.append(vals)
        df = pd.DataFrame(compiled_data[1:], columns=compiled_data[0]).apply(
            pd.to_numeric, errors="ignore"
        )
        if "lmiss" in file or "hwe" in file:
            return pd.concat(
                [
                    pd.DataFrame(
                        df["SNP"].str.split("_").to_list(),
                        columns=["ID", "CHROM", "POS", "REF", "ALT"],
                    ),
                    df.iloc[:, 2:],
                ],
                axis=1,
            )
        return df.rename(columns={"IID": "SAMPLE_ID"}).drop(columns="FID")

    def genotype_qc(self):
        raise NotImplementedError("Genotype level QC has not been implemented yet.")

    def variant_qc(self):
        if self.variant_call_rate:
            self.vc_file = self.filter_on_variant_call_rate()
        if self.hwe:
            self.hwe_file = self.filter_on_hwe()

    def sample_qc(self):
        self.sc_file = self.filter_on_sample_call_rate()

    @staticmethod
    def _compute_base_qc(file, output_prefix):
        # Computes QC metrics. Need to be recomputed after each sequential metric
        run_command(
            [
                "plink",
                "--vcf",
                file,
                "--hardy",
                "midp",
                "--missing",
                "--out",
                output_prefix,
            ]
        )

    def _compute_variant_call_rate(self, file_input):
        output_prefix = os.path.join(self.dir_, "vcr")
        self._compute_base_qc(file=file_input, output_prefix=output_prefix)
        vcr = (
            self._parse_metrics_file(f"{output_prefix}.lmiss")
            .assign(VARIANT_CALL_RATE=lambda x: 1 - x["F_MISS"])
            .drop(columns="F_MISS")
        )
        vcr.to_csv(
            os.path.join(self.dir_, "variant_call_rate.tsv"), sep="\t", index=False
        )

    def _compute_hwe(self, file_input):
        output_prefix = os.path.join(self.dir_, "hwe")
        self._compute_base_qc(file=file_input, output_prefix=output_prefix)
        hwe = self._parse_metrics_file(f"{output_prefix}.hwe").drop(
            columns=["TEST", "A1", "A2"]
        )
        hwe.to_csv(os.path.join(self.dir_, "hwe.tsv"), sep="\t", index=False)

    def _compute_sample_call_rate(self, file_input):
        output_prefix = os.path.join(self.dir_, "scr")
        self._compute_base_qc(file=file_input, output_prefix=output_prefix)
        scr = (
            self._parse_metrics_file(f"{output_prefix}.imiss")
            .assign(SAMPLE_CALL_RATE=lambda x: 1 - x["F_MISS"])
            .drop(columns="F_MISS")
        )
        scr.to_csv(
            os.path.join(self.dir_, "sample_call_rate.tsv"), sep="\t", index=False
        )

    @index_file
    @log(logger=logger)
    def filter_on_variant_call_rate(self):
        file_input = self.output_file
        # Compute metrics file
        self._compute_variant_call_rate(file_input=file_input)

        # Filter on variant call rate
        vc_qc_file = os.path.join(self.dir_, "vc_qced.vcf.gz")
        run_command(
            f"bcftools view --include 'F_MISSING < {1-self.variant_call_rate}' -Oz -o {vc_qc_file} {self.file}",
            shell=True,
        )
        return vc_qc_file

    @index_file
    @log(logger=logger)
    def filter_on_hwe(self):
        # Computes HWE using mid p-value correction
        file_input = self.output_file

        # Compute metrics file
        self._compute_hwe(file_input=file_input)
        hwe_qc_file = os.path.join(self.dir_, "hwe_qced")
        run_command(
            [
                "plink",
                "--vcf",
                file_input,
                "--hwe",
                str(self.hwe),
                "midp",
                "--out",
                hwe_qc_file,
            ]
            + self.plink_vcf_flags
        )
        return f"{hwe_qc_file}.vcf.gz"

    @index_file
    @log(logger=logger)
    def filter_on_sample_call_rate(self):
        file_input = self.output_file
        self._compute_sample_call_rate(file_input=file_input)

        sc_qc_file = os.path.join(self.dir_, "sc_qced")
        run_command(
            [
                "plink",
                "--vcf",
                file_input,
                "--mind",
                str(1 - self.sample_call_rate),
                "--out",
                sc_qc_file,
            ]
            + self.plink_vcf_flags
        )
        return f"{sc_qc_file}.vcf.gz"

    @property
    def output_file(self):
        for attr in self.attrs_created:
            try:
                file = getattr(self, attr)
            except AttributeError:
                pass
            else:
                return file
        return self.file

    @property
    def files_created(self):
        files_created = []
        for file in self.attrs_created:
            try:
                attr = getattr(self, file)
            except AttributeError:
                pass
            else:
                files_created.append(attr)
        return files_created



from setuptools import setup, find_packages

long_description = open("README.md").read()

setup(
    name="pharmcat_runner",
    version="0.0.15",
    description="Installs, runs, and parsers PharmCAT using Pharmacoscan input.",
    author="Andrew Haddad, PharmD",
    author_email="andrew.haddad@pitt.edu",
    packages=find_packages(),
    long_description=long_description,
    long_description_content_type="text/markdown",
    entry_points={
        "console_scripts": ["pharmcat_runner = pharmcat_runner.__main__:main"]
    },
    install_requires=["pandas"],
    package_data={"pharmcat_runner": ["lib/*"]},
    include_package_data=True,
)
